{
 "cells": [
  {
   "cell_type": "markdown",
   "id": "274986cf-eb9d-43fb-9e5e-747375967074",
   "metadata": {},
   "source": [
    "# Synchronization of Threads Using Barrier and Ordered Directive"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "47ec2fe5-66c0-40e3-aeb2-128ddef90823",
   "metadata": {
    "jp-MarkdownHeadingCollapsed": true,
    "tags": []
   },
   "source": [
    "## Introduction\n",
    "\n",
    "OpenMP provides a selection of tools to achieve synchronization between threads, ensuring smooth collaboration and preventing the pitfalls of concurrent execution. Among these, the `barrier` and `ordered` directives stand as pillars of thread coordination, each playing a distinct role in maintaining order and predictability within parallel regions.\n",
    "\n",
    "- **Barrier Directive**: Acting as a synchronization point, the barrier directive ensures all threads within a team reach a designated point before any are allowed to proceed. This is crucial in preventing race conditions, where multiple threads accessing and modifying shared data simultaneously can lead to unpredictable and erroneous results.\n",
    "\n",
    "- **Ordered Directive**: For scenarios where the order of execution is critical, the ordered directive ensures specific sections within a parallelized loop are executed sequentially, respecting the order of iterations. This is essential for maintaining deterministic behavior in situations like I/O operations or when data dependencies exist between iterations."
   ]
  },
  {
   "cell_type": "markdown",
   "id": "6fb8c339-d224-4e79-b07e-2b29979c3aa4",
   "metadata": {},
   "source": [
    "## Barrier Directive\n",
    "\n",
    "The `barrier` directive acts as a crucial tool for orchestrating thread synchronization within a parallel region. It establishes designated points where all threads within a team must converge before any are permitted to proceed. This mechanism is vital for preventing race conditions and maintaining data consistency in shared-memory parallel programs, ensuring predictable and reliable results.\n",
    "\n",
    "### Establishing Synchronization Points\n",
    "\n",
    "The syntax for employing the barrier directive is straightforward and consistent across both C/C++ and Fortran:\n",
    "\n",
    "**C/C++:**\n",
    "\n",
    "```\n",
    "#pragma omp barrier\n",
    "```\n",
    "\n",
    "**Fortran:**\n",
    "\n",
    "```\n",
    "!$omp barrier\n",
    "```\n",
    "\n",
    "By strategically placing this directive within a parallel region, you create explicit synchronization points. Let's delve into a concrete example to illustrate this:\n",
    "\n",
    "```\n",
    "#pragma omp parallel\n",
    "{\n",
    "  // Each thread performs its independent calculations on a portion of data...\n",
    "  #pragma omp barrier\n",
    "  // All threads synchronize here, ensuring consistent data before the next step\n",
    "  // Combine the partial results from each thread into a final result...\n",
    "}\n",
    "```\n",
    "\n",
    "In this scenario, each thread works on its assigned portion of data before encountering the barrier. The `barrier` directive acts as a gatekeeper, halting each thread until all threads within the team have reached this point. This guarantees that all partial calculations are complete and the shared data is in a consistent state before moving on to combine the partial results.\n",
    "\n",
    "### Example"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "0f86ad17-a4b0-411d-9921-c05186c93225",
   "metadata": {},
   "outputs": [],
   "source": [
    "//%compiler: clang\n",
    "//%cflags: -fopenmp\n",
    "\n",
    "#include <omp.h>\n",
    "#include <stdio.h>\n",
    "\n",
    "int main() {\n",
    "    int sum = 0;\n",
    "    #pragma omp parallel shared(sum)\n",
    "    {\n",
    "        int tid = omp_get_thread_num();\n",
    "        int local_sum = 0;\n",
    "        // Each thread calculates a local sum for a portion of data\n",
    "        for (int i = tid * 10; i < (tid + 1) * 10; ++i) {\n",
    "            local_sum += i;\n",
    "        }\n",
    "        \n",
    "        #pragma omp critical\n",
    "        sum += local_sum; // Add each thread's local sum to the global sum\n",
    "        \n",
    "        #pragma omp barrier // All threads wait here before printing\n",
    "        \n",
    "        printf(\"Thread %d finished with local sum %d\\n\", tid, local_sum);\n",
    "    }\n",
    "    printf(\"Final sum: %d\\n\", sum);\n",
    "    return 0;\n",
    "}"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "85d34895-5cfd-49c9-99f0-a8d2c5f49b7d",
   "metadata": {},
   "source": [
    "## The Ordered Directive: Maintaining Sequential Steps\n",
    "\n",
    "In the realm of parallelized loops, where iterations dance across threads, maintaining a specific execution order for certain sections of code becomes essential. The ordered directive steps onto the stage, ensuring that designated portions of a loop are executed sequentially, respecting the order of iterations even when the loop itself is parallelized and iterations might be executed concurrently.\n",
    "\n",
    "### Enforcing Order in the Parallel Ballet\n",
    "\n",
    "The syntax for invoking the ordered directive is clear and consistent across C/C++ and Fortran:\n",
    "\n",
    "**C/C++:**\n",
    "\n",
    "```\n",
    "#pragma omp ordered\n",
    "structured-block\n",
    "#pragma omp end ordered\n",
    "```\n",
    "\n",
    "**Fortran:**\n",
    "```\n",
    "!$omp ordered\n",
    "structured-block\n",
    "!$omp end ordered\n",
    "```\n",
    "\n",
    "Let's explore a simple example to grasp the essence of the ordered directive:\n",
    "\n",
    "```\n",
    "#pragma omp parallel for ordered\n",
    "for (int i = 0; i < N; ++i) {\n",
    "  // Perform calculations specific to iteration i...\n",
    "  #pragma omp ordered\n",
    "  printf(\"Iteration %d from thread %d\\n\", i, omp_get_thread_num());\n",
    "}\n",
    "```\n",
    "\n",
    "In this scenario, even though the loop iterations might be scattered across threads and executed concurrently, the printf statement within the ordered region will always print in the order of the loop iterations (0, 1, 2, ...). The ordered directive guarantees that each thread executes this section sequentially, respecting the natural order of the loop.\n",
    "### doacross Clause: Specifying Dependencies\n",
    "For more intricate scenarios involving loops with cross-iteration dependencies, the doacross clause comes to the rescue. It allows you to pinpoint specific dependences between iterations that must be honored during parallel execution. By using the doacross clause, you provide explicit instructions to the compiler regarding the required order of operations, ensuring correctness even in the presence of complex data dependencies."
   ]
  },
  {
   "cell_type": "markdown",
   "id": "8708a7ee-6f10-4f9f-a36a-709d6eb7c652",
   "metadata": {},
   "source": [
    "## Implicit Barriers: Automatic Synchronization\n",
    "\n",
    "OpenMP provides automatic synchronization with implicit barriers placed at the end of worksharing constructs (like for and sections) and at the end of parallel regions. These invisible barriers guarantee that all threads have finished their assigned work within the construct before moving on to the next section of code.\n",
    "**Example:**\n",
    "```\n",
    "#pragma omp parallel for\n",
    "for (int i = 0; i < N; ++i) {\n",
    "    // Work on iteration i...\n",
    "}\n",
    "// Implicit barrier - all threads wait here before proceeding\n",
    "```\n",
    "\n",
    "### Removing Implicit Barriers with nowait\n",
    "\n",
    "While implicit barriers provide convenient synchronization, they might introduce unnecessary overhead in certain scenarios. The nowait clause allows you to remove these implicit barriers, enabling threads to proceed without waiting for others. However, use this clause judiciously and ensure proper explicit synchronization mechanisms are in place to avoid race conditions.\n",
    "\n",
    "**Example:**\n",
    "\n",
    "```\n",
    "#pragma omp parallel for nowait\n",
    "for (int i = 0; i < N; ++i) {\n",
    "    // Work on iteration i...\n",
    "}\n",
    "// No implicit barrier - threads continue immediately\n",
    "```\n"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "5157b14f-230e-4734-8465-2746b75df34e",
   "metadata": {},
   "source": [
    "## Best Practices for Using Barrier and Ordered Directives: Achieving Synchronization Zen\n",
    "While the barrier and ordered directives are powerful tools for orchestrating thread synchronization, their effectiveness hinges on understanding their nuances and applying them strategically. Here are some guiding principles to help you achieve synchronization zen and unlock the full potential of these directives:\n",
    "- Barrier Mindfulness: Barriers, while essential for preventing race conditions, can introduce overhead due to thread idling and synchronization costs. Use them judiciously and only when necessary to ensure data consistency and correctness. Explore alternative synchronization mechanisms like locks or atomic operations when appropriate.\n",
    "- Embrace Implicit Barriers: Leverage the convenience and efficiency of implicit barriers automatically inserted at the end of worksharing constructs and parallel regions. Be mindful of their presence and consider the nowait clause only when you are confident about the absence of race conditions and its impact on correctness.\n",
    "- Order with Purpose: Apply the ordered directive selectively to code sections within a loop where maintaining the order of execution is critical. Overusing the ordered directive can diminish the benefits of parallelization, so carefully analyze your code's data dependencies and apply it only when necessary.\n",
    "- Data Dependencies Awareness: Gain a deep understanding of the data dependencies within your code, both within iterations and across iterations. This knowledge is crucial for selecting the appropriate synchronization directives and ensuring that threads access and modify shared data in a controlled manner.\n",
    "- Performance as the Guiding Light: Employ performance profiling tools to identify and address potential bottlenecks arising from synchronization overhead. Optimize your code by minimizing unnecessary barriers and restructuring your loops to reduce cross-iteration dependencies, striking a balance between synchronization and parallel efficiency."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "eeec29df-f6d3-4fc8-9720-8955a3cb09e6",
   "metadata": {},
   "outputs": [],
   "source": []
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Native",
   "language": "native",
   "name": "native"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.8.10"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}
